/*
******************************************************************************
**  IO Interface to the Typhoon System
**
**  - used for CarMaker Office and HIL 
**  - system is connected by CAN
**
**  --------------------------------------------------------------------------
**  Copyright (C)  ReitzeTechnologies
**                 Elfenweg 10             eMail  Info@Reitze-Technologies.de
**                 D-76199 Karlsruhe
**                 Germany
******************************************************************************
*/


#include <Global.h>

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <string.h>


#include <CarMaker.h>

#if defined(CM_HIL)
# include <mio.h>
# include <can_interface.h>
# include <FailSafeTester.h>
# if !defined(LABCAR)
# include <flex.h>
# include <rbs.h>
# endif
# include <CM_XCP.h>
# include <CM_CCP.h>
#endif // of #if !defined(CM_HIL)


#if !defined(CM_HIL)
#  include <pcanio.h>
#endif

#include <Car/Vehicle_Car.h>
#include "IO_CM2Ty.h"


/* Supported IO Configurations: */
unsigned int IO_CM2Ty_MIO =  0;	// not yet supported
unsigned int IO_CM2Ty_PCAN = 0;	// CM office: by PCAN USB 2 CAN module


tIO_CM2Ty IO_CM2Ty;


/*
** send_can_message ()
**
** Send CAN message by PCAN USB interface
*/

static void
send_can_message (int msg_id, void* val) {

    CAN_Msg m;

    m.MsgId = msg_id;
    m.FrameFmt = 0;
    m.RTR = 0;
    m.FrameLen = 4;
    memcpy(m.Data, val, 4);

    if (IO_CM2Ty_PCAN) {
	PCANIO_Send(pcan_usb, 0, &m);
    }
}



/*
** send_two_vals_can_message ()
**
** Send CAN message by PCAN USB interface
*/

static void
send_two_vals_can_message(int msg_id, void* val1, void* val2) {

    static CAN_Msg m = { } ;

    m.MsgId = msg_id;
    m.FrameFmt = 0;
    m.RTR = 0;
    m.FrameLen = 8;
    memcpy(m.Data,   val1, 4);
    memcpy(m.Data+4, val2, 4);

    if (IO_CM2Ty_PCAN) {
	PCANIO_Send(pcan_usb, 0, &m);
    }
}




/*
** IO_CM2Ty_Init_First ()
**
** First, low level initialization of the IO module
**
** Call:
** - one times at start of program
** - no realtime conditions
*/

int
IO_CM2Ty_Init_First (void)
{
    memset(&IO_CM2Ty,  0, sizeof(IO_CM2Ty));

    //IO_CM2Ty_SetConfigurations(IOConfiguration);

    /*** initialize the IO modules for this connection ***/
    /* @@@ */
    /* ... */
    
    return 0;
}



/*
** IO_CM2Ty_Init ()
**
** initialization
** - i/o hardware
** - add variables to data dictionary
**
** call:
** - single call at program start
*/

int
IO_CM2Ty_Init (void)
{
    Log ("Function %s () ...\n", __FUNCTION__);
    struct tIO_CM2Ty *io =  &IO_CM2Ty;
    tDDefault *df = DDefaultCreate("IO.CM2Ty.");
    tDDictEntry *e = NULL;
    int rv = 0;

    if (IO_CM2Ty_PCAN) {
	/*** office version */
	int dev = pcan_usb;
	int ch = 0;
	int Baudrate = 1000000;
	char msgbuf[1024] = "";
		
	rv = PCANIO_Init();
	Log ("PCAN initialize, returns %d\n", rv);

	rv = PCANIO_GetStatus(dev, ch, msgbuf);
	Log ("PCAN GetStatus returns %d\n", rv);

	PCANIO_SetCommParam(dev, ch, Baudrate, 0, NULL, NULL);
    }

    if (IO_CM2Ty_MIO || IO_CM2Ty_PCAN) {
	/*** Com */
	e = DDefUChar (df, "Com.TyActive",	"", 	&io->Com.TyActive,     DVA_IO_In);
	DDefStates(e, 2, 0);
	
	e = DDefUChar (df, "Com.TyRqst",	"", 	&io->Com.TyRqst,       DVA_IO_Out);
	DDefStates(e, 2, 0);

	
	/*** VhclCtrl */
	e = DDefInt   (df, "VC.Key",		"", 	&io->VhclCtrl.Key,       DVA_IO_Out);
	DDefStates(e, 5, 0);

	e = DDefFloat (df, "VC.Gas",		"", 	&io->VhclCtrl.Gas,       DVA_IO_Out);
	e = DDefFloat (df, "VC.Brake",		"", 	&io->VhclCtrl.Brake,     DVA_IO_Out);
	e = DDefFloat (df, "VC.Load",		"", 	&io->VhclCtrl.Load,      DVA_IO_Out);


	/*** Motor0 */
	struct tIO_CM2Ty_Motor *m = NULL;

	m = &io->Motor[0];
	e = DDefUChar (df, "M0.Active",		"", 	&m->Active,       DVA_IO_Out);
	DDefStates(e, 2, 0);

	e = DDefFloat (df, "M0.Trq",		"Nm", 	&m->Trq,       	  DVA_IO_In);
	e = DDefFloat (df, "M0.Load",		"",	&m->Load,      	  DVA_IO_Out);
	e = DDefFloat (df, "M0.Spd",		"rad/s",&m->Spd,       	  DVA_IO_Out);

	/*** Motor1 */
	m = &io->Motor[1];
	e = DDefUChar (df, "M1.Active",		"", 	&m->Active,       DVA_IO_Out);
	DDefStates(e, 2, 0);

	e = DDefFloat (df, "M1.Trq",		"Nm", 	&m->Trq,       	  DVA_IO_Out);
	e = DDefFloat (df, "M1.Load",		"",	&m->Load,      	  DVA_IO_Out);
	e = DDefFloat (df, "M1.Spd",		"rad/s",&m->Spd,       	  DVA_IO_In);
    }								  

    if (IO_CM2Ty_MIO) {
	/*@@@*/
	/*@@@ add varaibles to data dictionary */
    }

    if (IO_CM2Ty_PCAN) {
	/*@@@*/
	/*@@@ add varaibles to data dictionary */
    }
    
    DDefaultDelete (df);					
    return 0;
}



/*
** IO_CM2Ty_Init_Finalize ()
**
** @@@ not yet used @@@
**
** last (deferred) I/O initialization step
**
** call:
** - single call at program start in CarMaker_FinishStartup()
*/

int
IO_CM2Ty_Init_Finalize (void)
{
    if (IO_CM2Ty_MIO) {
	/*@@@*/
    }

    if (IO_CM2Ty_PCAN) {
	/*@@@*/
#if defined (TYMOTOR_DEBUG)
	IO_CM2Ty.Com.TyActive = 1; //should be transmitted by Typhoon
	Log ("Set Typhoon communication to activ manually (%s:%d)\n", __FILE__, __LINE__);
#endif // of #if defined (TYMOTOR_DEBUG)
	
    }
    
    return 0;
}



/*
** IO_CM2Ty_Param_Get ()
**
** Get i/o configuration parameters
** - calibration
** - constant values
** - ids
*/

int
IO_CM2Ty_Param_Get (tInfos *inf)
{
    unsigned nError = GetInfoErrorCount ();

    /* ignition off */
    SetKl15 (0);

    if (IO_None)
    	return 0;

    if (IO_CM2Ty_MIO || IO_CM2Ty_PCAN) {
	/*@@@*/
    }

    if (IO_CM2Ty_MIO) {
	/*@@@*/
    }

    if (IO_CM2Ty_PCAN) {
	/*@@@*/
    }

    return nError != GetInfoErrorCount() ? -1 : 0;
}





/*
** IO_CM2Ty_In ()
**
** reading signals from hardware / ECU
**
** CycleNo: simulation cycle counter, incremented every loop/millisecond
**
** call:
** - in the main loop
** - first function call in main loop, after waiting for next loop
** - just before User_In()
** - pay attention to realtime condition
*/

void
IO_CM2Ty_In (unsigned CycleNo)
{

    if (IO_None)
	return;

    if (IO_CM2Ty_MIO) {
	/* not yet implemented @@@*/
    }

    if (IO_CM2Ty_PCAN) {
	CAN_Msg Msg;
	while (PCANIO_Recv(pcan_usb, 0, &Msg) == 0) {
	    int ival1;
	    int ival2;

	    IO_CM2Ty.Com.TyActive = 1; //should be transmitted by Typhoon

	    //struct tIO_CM2Ty_Motor *mif = NULL;
	    switch (Msg.MsgId) {

	      case 701:
		//mif = &IO_CM2Ty.Motor[0];
		memcpy (&ival1, &Msg.Data[0], 4);
		memcpy (&ival2, &Msg.Data[4], 4);
		//IO_CM2Ty.Motor[0].Spd = ival1;		//speed
		IO_CM2Ty.Motor[0].Trq = ival2*0.01;		//torque
		break;

	      case 801: 
		memcpy (&ival1, &Msg.Data[0], 4);
		memcpy (&ival2, &Msg.Data[4], 4);
		//IO_CM2Ty.Motor[1].Spd = ival1;				//speed
		IO_CM2Ty.Motor[1].Trq = ival2*0.01;		//torque
		break;

	      case 999:
		memcpy (&ival1, &Msg.Data[0], 4);
		memcpy (&ival2, &Msg.Data[4], 4);
		//IO_CM2Ty.Motor[1].Spd = ival1;				//speed
		//IO_CM2Ty.Motor[1].Trq = ival2*0.01;		//torque
		break;

	      default:
		break;
	    }
	}
    }
}



/*
** IO_CM2Ty_Out ()
**
** writing signals to hardware / ECU
**
** CycleNo: simulation cycle counter, incremented every loop/millisecond
**
** call:
** - in the main loop
** - last function call in main loop
** - just after User_Out()
** - pay attention to realtime condition
*/

void 
IO_CM2Ty_Out (unsigned CycleNo) 
{
    if (IO_None)
	return;

    if (IO_CM2Ty_MIO) {
	/* not yet implemented @@@*/
    }

    if (IO_CM2Ty_PCAN && IO_CM2Ty.Com.TyRqst) {

	if (IO_CM2Ty.Com.TyParamRqst) {
	    int val01 = 1;
	    send_can_message(901, &val01);
	    IO_CM2Ty.Com.TyParamRqst = 0;
	}
	
	// General
	// send each 10th cycle
	if (CycleNo%10 == 1) {
	    int val01 = Car.ConBdy1.v * 1000;
	    int val02 = VehicleControl.Gas*1000;
	    int val03 = VehicleControl.Brake*1000;
	    send_can_message (101, &val01);
	    send_two_vals_can_message (102, &val02, &val03);
	    //send_can_message(103, &val03);
	}

	// Motor0
	// send each 10th cycle
	if (CycleNo%10 == 2) {
	    int val11 = PowerTrain.MotorIF[0].Trq * 1000;
	    int val12 = PowerTrain.MotorIF[0].rotv; 
	    //send_can_message(201, &val11);
	    //send_can_message(202, &val12);
	    send_two_vals_can_message(201, &val11, &val12);		
	}

	// Motor1
	// send each 10th cycle
	if (CycleNo%10 == 3) {
	    int val21 = PowerTrain.MotorIF[1].Trq * 1000;
	    int val22 = PowerTrain.MotorIF[1].rotv;
	    //send_can_message(301, &val21);
	    //send_can_message(302, &val22); 
	    send_two_vals_can_message(301, &val21, &val22); 
	}
	
	static float SOC_Send = 0; 
	static unsigned int SOC_Counter = 0;
	SOC_Send += PowerTrain.BatteryCU_IF.SOC_HV;
	SOC_Counter ++ ;
	
	if (CycleNo%1000 == 1) {
	    int val31 = (SOC_Send *100.0)  / SOC_Counter;
	    int val32 = 1;
	    //Log ("SOC Send %6d \n", val31); 	/*@@@*/
	    send_two_vals_can_message(401, &val31, &val32);
	    SOC_Send = 0;
	    SOC_Counter = 0;
	}
    }
}



/*
** IO_CM2Ty_Cleanup ()
**
** Uninits all MIO hardware:
** - puts M-Modules into reset state
** - frees unneeded memory
*/

void
IO_CM2Ty_Cleanup (void)
{
    if (IO_None)
	goto EndReturn;

    if (IO_CM2Ty_MIO) {
	/*@@@*/
    }

    if (IO_CM2Ty_PCAN) {
	/*@@@*/
    }

  EndReturn:
    return;
}
