/*
******************************************************************************
**
** Motor Model interface to a E-Motor running on a Typhoon HIL system
**
** To use this module, add this model to CarMakers model library by calling
**
**    Motor_Register_TyMotor0 ()
**
******************************************************************************
*/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <SimCore.h>
#include <CarMaker.h>
#include <Car/Vehicle_Car.h>
#include <VehicleControl.h>

#include "IO_CM2Ty.h"

static const char ThisModelClass[] = "Powertrain.Motor";
static const char ThisModelKind[]  = "TyMotor0";
static const int  ThisVersionId    = 1;

static unsigned int   MotorIndex = 0;     // index for Motor[]
#define MsgPreStr     "TyMotor0: "
#define REGISTER_FCN  Motor_Register_TyMotor0


struct tMyModel {
    /* paramters */
    double Inert;
    double Ratio;
    double Trq_max;	//@@@ without PCAN, debugging

    /* signals */
    double rotv;
    double Trq;

    unsigned int Wait4TyActive;
};

static void MyModel_Delete (void *MP);



static void
MyModel_DeclQuants_dyn (struct tMyModel *mp, int park)
{
    static struct tMyModel MyModel_Dummy = {0};
    if (park) {
        mp = &MyModel_Dummy;
    }
    
    /* Define here dict entries for dynamically allocated variables. */
    char buf[128];

    sprintf (buf, "PT.%s.", ThisModelKind);
    tDDefault *df = DDefaultCreate(buf);
    tDDictEntry *e;

    //@@@ DVA_PTC ok?
    e = DDefDouble (df, "rotv",         "rad/s",        &mp->rotv,          DVA_PTC);
    e = DDefDouble (df, "Trq",          "Nm",           &mp->Trq,           DVA_PTC);
    
    e = DDefUInt  (df, "Wait4TyActive", "",             &mp->Wait4TyActive, DVA_PTC);
    DDefStates(e, 0, 0);
    
    DDefaultDelete (df);                                            
}



static void
MyModel_DeclQuants (void *MP)
{
    struct tMyModel *mp = (struct tMyModel *)MP;

    if (mp == NULL) {
        /* Define here dict entries for non-dynamically allocated (static) variables. */

    } else {
        MyModel_DeclQuants_dyn (mp, 0);
    }
}


/*
** Model output parameters in the configuration struct CfgIF, which are required
** by CarMaker, are read in before the MyModel_New() function.
** - The parametrization of these parameters is supported by the GUI.
** - These output parameters can be used internally by the model in same way like
**   the input parameters
*/

static void *
MyModel_New (
    struct tInfos *Inf,
    struct tPTMotorCfgIF *CfgIF,
    const char *KindKey,
    const char *Ident)
{
    struct tMyModel *mp = NULL;
    const char *ModelKind;
    char MsgPre[64];
    int VersionId = 0;
    unsigned int IErrCount = GetInfoErrorCount();
    int i;

    sprintf (MsgPre, "%s %s", ThisModelClass, ThisModelKind);

    if ((ModelKind = SimCore_GetKindInfo(Inf, ModelClass_PTMotor, KindKey,
                                         0, ThisVersionId, &VersionId)) == NULL)
        return NULL;

    mp = (struct tMyModel*)calloc(1,sizeof(*mp));

    /* get CfgIF parameters */
    if (Motor_GetCfgOutIF (Inf, CfgIF, ModelKind, Ident) != 0)
        goto ErrorReturn;

    /* CfgIF output -> Model */
    mp->Ratio  = CfgIF->Ratio;
    mp->Ratio  = iGetDbl(Inf, "PowerTraion.Motor.Ratio");

    mp->Inert  = iGetDbl(Inf, "PowerTraion.Motor.Inert");
    mp->Inert *= mp->Ratio * mp->Ratio;

    mp->Trq_max  = iGetDbl(Inf, "PowerTraion.Motor.Trq_max");
    
#if 0
    /* CfgIF output: verification if the parametrization corresponds to the model */
    if (CfgIF->TrqMot_max==NULL) {
        LogErrF (EC_Init, "%s: no characteristic for motor torque", MsgPre);
        goto ErrorReturn;
    }
    //CfgIF->TrqGen_max in the model ignored
#endif


    Log ("%s: rotv_mot_max = %8.3f \n", MsgPre, CfgIF->rotv_Mot_max);
    Log ("%s: rotv_gen_max = %8.3f \n", MsgPre, CfgIF->rotv_Gen_max);
    Log ("%s: Ratio = %+8.5f \n",       MsgPre, CfgIF->Ratio);

    i = GetInfoErrorCount();
    if (IErrCount != i) {
        LogErrF (EC_Init, "%s: Errors occured while reading parameters", MsgPre);
        goto ErrorReturn;
    }
    
    struct tIO_CM2Ty       *io =  &IO_CM2Ty;
    struct tIO_CM2Ty_Motor *iom = &IO_CM2Ty.Motor[MotorIndex];

    io->Com.TyRqst =      1;
    io->Com.TyParamRqst = 1;    //@@@ for one or two motors?
    iom->Active =         1;    //@@@ ok?

    /* @@@ wait for n seconds that typhoon answers with TyActive=1 and parameter messages */

    return mp;
    
    ErrorReturn:
    MyModel_Delete (mp);
    return NULL;
}



/*
** MyModel_Calc ()
**
**
** - this function is called each PowerTrain integration substep
**
*/

static int
MyModel_Calc (void *MP, struct tPTMotorIF *IF, double dt)
{
    struct tMyModel *mp = (struct tMyModel *)MP;
    //const tPTMotorCfgIF *CfgIF = IF->CfgIF;

    struct tIO_CM2Ty       *io =  &IO_CM2Ty;
    struct tIO_CM2Ty_Motor *iom = &IO_CM2Ty.Motor[MotorIndex];
    
    if (io->Com.TyActive != 1) {
        mp->Wait4TyActive++;
        /*@@@ define limit, error handling */
#if defined (TYMOTOR_DEBUG)
        Log ("%sTyphoon connection not active, n=%d\n", MsgPreStr, mp->Wait4TyActive);
#endif // of #if defined (TYMOTOR_DEBUG)

        return 0;
    }
    
    /*** TyActive == 1 */
    mp->Wait4TyActive = 0;
    io->Com.TyRqst =    1;
    
    mp->rotv =  IF->rotv * mp->Ratio;
    iom->Spd =  mp->rotv;       // to IO
    iom->Load = IF->Load;
    
    //mp->Trq  = LMEval (CfgIF->TrqMot_max, fabs(mp->rotv)) * IF->Load;
    mp->Trq  = iom->Trq;        // from IO

        
    IF->Trq =     mp->Trq * mp->Ratio;

#if defined (TYMOTOR_DEBUG)
    IF->Trq =     1000.0*VehicleControl.Gas;
#endif

    IF->PwrElec = IF->Trq * IF->rotv;
    IF->Inert =   mp->Inert;


#if defined (DEBUG)
    if(SimCore.CycleNo%2000 == 1) {
        Log("%s%5d rotv=%8.3f  Trq=%5.2f  Load=%5.3f VC.Gas=%5.3f\n",
	    MsgPreStr,
	    SimCore.CycleNo,
	    iom->Spd,
	    IF->Trq,
	    IF->Load, VehicleControl.Gas);
    }
#endif // of #if defined (DEBUG)
    
    return 0;
}




static void
MyModel_Delete (void *MP)
{
    struct tMyModel *mp = (struct tMyModel *)MP;
    struct tIO_CM2Ty       *io =  &IO_CM2Ty;
    struct tIO_CM2Ty_Motor *iom = &IO_CM2Ty.Motor[MotorIndex];

    io->Com.TyRqst = 0;
    iom->Active =    0;
    
    /* Park the dict entries for dynamically allocated variables before deleting */
    MyModel_DeclQuants_dyn (mp, 1);
    free (mp);
}



int 
REGISTER_FCN (void)
{
    tModelClassDescr m;

    memset(&m, 0, sizeof(m));
    m.PTMotor.VersionId =       ThisVersionId;
    m.PTMotor.New =             MyModel_New;
    m.PTMotor.Calc =            MyModel_Calc;
    m.PTMotor.Delete =          MyModel_Delete;
    m.PTMotor.DeclQuants =      MyModel_DeclQuants;
    /* Should only be used if the model doesn't read params from extra files */
    m.PTMotor.ParamsChanged =   ParamsChanged_IgnoreCheck;

    return Model_Register(ModelClass_PTMotor, ThisModelKind, &m);
}
